﻿// lab_1_basics.cpp : Определяет точку входа для приложения.


#include <Windows.h>
#include <commctrl.h>
#include "stdio.h"
#include "frame.hpp"
#include "painter.hpp"

// Windows-приложение для создания буфера кадра

LRESULT CALLBACK WindowProc(HWND, UINT, WPARAM, LPARAM);

HWND hWndStatusBar; // Дескриптор компонента StatusBar

int WINAPI WinMain(HINSTANCE hThisInstance, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode) {
    char szWinName[] = "Graphics Window Class"; // Имя класса окна

    HWND hWnd; // Дескриптор главного окна
    MSG msg;
    WNDCLASSA wcl; // Определитель класса окна
    wcl.hInstance = hThisInstance; // Дескриптор приложения
    wcl.lpszClassName = szWinName;// Имя класса окна
    wcl.lpfnWndProc = WindowProc; // Функция обработки сообщений
    wcl.style = 0; // Стиль по умолчанию
    wcl.hIcon = LoadIcon(nullptr, IDI_APPLICATION);// Иконка
    wcl.hCursor = LoadCursor(nullptr, IDC_ARROW); // Курсор
    wcl.lpszMenuName = nullptr; // Без меню
    wcl.cbClsExtra = 0; // Без дополнительной информации
    wcl.cbWndExtra = sizeof(PainterState *) + sizeof(Painter **);
    static_assert(sizeof(PainterState *) == sizeof(LONG_PTR));
    static_assert(sizeof(Painter **) == sizeof(LONG_PTR));


    wcl.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH); //Белый фон

    if (!RegisterClassA(&wcl)) // Регистрируем класс окна
        return 0;

    hWnd = CreateWindowA(szWinName, // Создать окно
                         "Лабораторная работа №1. Буфер кадра. Алгоритмы Брезенхейма",
                         WS_OVERLAPPEDWINDOW, // Стиль окна
                         CW_USEDEFAULT, // x-координата
                         CW_USEDEFAULT, // y-координата
                         CW_USEDEFAULT, // Ширина
                         CW_USEDEFAULT, // Высота
                         HWND_DESKTOP, // Без родительского окна
                         nullptr, // Без меню
                         hThisInstance, // Дескриптор приложения
                         nullptr); // Без дополнительных аргументов

    // Создаём компонент типа StatusBar
    hWndStatusBar = CreateWindowExA(
            0, STATUSCLASSNAMEA, nullptr,
            WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP,
            0, 0, 0, 0,
            hWnd, (HMENU) 10001,
            hThisInstance, nullptr
    );

    // Настройка частей StatusBar'а
    int statwidths[] = {150, 300, -1};
    SendMessageA(hWndStatusBar, SB_SETPARTS, sizeof(statwidths) / sizeof(int), (LPARAM) statwidths);

    ShowWindow(hWnd, nWinMode); // Показать окно
    UpdateWindow(hWnd); // Перерисовать окно

    while (GetMessage(&msg, nullptr, 0, 0)) // Запустить цикл обработки сообщений
    {
        TranslateMessage(&msg); // Разрешить использование клавиатуры
        DispatchMessage(&msg); // Вернуть управление операционной системе Windows
    }

    return msg.wParam;

}

// Следующая функция вызывается операционной системой Windows и получает в качестве
// параметров сообщения из очереди сообщений данного приложения
LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    static int pixelSize = 8; // Размер "большого" пикселя

    switch (message) {
        case WM_CREATE:
            SetTimer(hWnd, 1, 1000 / 30, nullptr);
            SetWindowLongPtr(hWnd, 0, (LONG_PTR) new PainterState());
            SetWindowLongPtr(hWnd, sizeof(PainterState *), (LONG_PTR) &predefined_painters[0]);
            break;

        case WM_PAINT: {
            PAINTSTRUCT ps;

            HDC hdc = BeginPaint(hWnd, &ps);

            // Определяем ширину и высоту окна
            RECT rect = ps.rcPaint;
            GetClientRect(hWnd, &rect);

            int width = rect.right - rect.left;
            int height = rect.bottom - rect.top;

            // Рисование в буфер кадра

            int ratio = pixelSize; // Размер "большого" пикселя

            int W = width / ratio;
            int H = (height - 22) / ratio; // Отнимем высоту StatusBar'а

            Frame frame(W, H);
            auto *state = reinterpret_cast<PainterState *>(GetWindowLongPtr(hWnd, 0));
            auto *painter = *reinterpret_cast<Painter **>(GetWindowLongPtr(hWnd, sizeof(PainterState *)));
            PixelGridPainter{COLOR{245, 245, 245}, COLOR{240, 240, 240}}.draw(state, &frame);
            painter->draw(state, &frame);

            // Системная структура для хранения цвета пикселя
            // Буфер кадра, который будет передаваться операционной системе, должен состоять из массива этих структур
            // Она не совпадает с порядком следования цветов в формате RBG
            typedef struct tagRGBPIXEL {
                unsigned char BLUE;        // Компонента синего цвета
                unsigned char GREEN;    // Компонента зелёного цвета
                unsigned char RED;        // Компонента красного цвета
                unsigned char ALPHA;    // Прозрачность
            } RGBPIXEL;

            // Выделение памяти для второго буфера, который будет передаваться функции CreateBitmap для создания картинки
            RGBPIXEL *bitmap = (RGBPIXEL *) HeapAlloc(GetProcessHeap(), 0, width * height * sizeof(RGBPIXEL));

            // Копирование массива пикселей в соответствии с системным форматом пикселя и масштабирование картинки
            // W и H - ширина и высота изображения в буфере кадра
            // ratio - коэффициент масштабирования пикселей
            for (int y = 0; y < H * ratio; y++)
                for (int x = 0; x < W * ratio; x++) {
                    RGBPIXEL *pixel = bitmap + y * width + x;
                    COLOR color = frame.get_pixel(x / ratio, y / ratio);
                    pixel->RED = color.red;
                    pixel->GREEN = color.green;
                    pixel->BLUE = color.blue;
                    pixel->ALPHA = color.alpha;
                }


            // Получить дескриптор на новое растровое изображение
            HBITMAP hBitMap = CreateBitmap(width, height, 1, sizeof(RGBPIXEL) * 8, bitmap);

            // Освободить память, которую занимает буфер цвета
            HeapFree(GetProcessHeap(), 0, bitmap);

            // Создать в оперативной памяти контекст, совместимый с экранным контекстом, который мы используем, чтобы рисовать
            HDC srcHdc = CreateCompatibleDC(hdc);

            // Связать картинку с новым контекстом
            SelectObject(srcHdc, hBitMap);

            // Копировать содержимое из временного контекста srcHdc в основной контекст окна hdc
            BitBlt(
                    hdc,        // Основной контекст
                    0, 0,       // Координаты левого верхнего угла, от которого будет выполняться вставка
                    width,      // Ширина вставляемого изображения
                    height,     // Высота вставляемого изображения
                    srcHdc,     // Дескриптор временного контекста
                    0, 0,       // Координаты считываемого изображения
                    SRCCOPY);   // Параметры операции - копирование

            EndPaint(hWnd, &ps);

            // Удаление картинки из памяти
            DeleteObject(hBitMap);

            // Удаление временного контекста
            DeleteDC(srcHdc);
        }
            break;

        case WM_MOUSEMOVE: {
            char str[256];

            // Устанавливаем текст в разных частях StatusBar'а
            // Экранные координаты курсора мыши
            sprintf_s(str, "X = %d, Y = %d", LOWORD(lParam), HIWORD(lParam));
            SendMessageA(hWndStatusBar, SB_SETTEXTA, 2, (LPARAM) str);

            // Координаты пикселя в буфере кадра
            sprintf_s(str, "BX = %d, BY = %d", LOWORD(lParam) / pixelSize, HIWORD(lParam) / pixelSize);
            SendMessageA(hWndStatusBar, SB_SETTEXTA, 1, (LPARAM) str);

            sprintf_s(str, "Масштаб (F2/F3): %d", pixelSize);
            SendMessageA(hWndStatusBar, SB_SETTEXTA, 0, (LPARAM) str);
        }
            break;

        case WM_LBUTTONDOWN:
            reinterpret_cast<PainterState *>(GetWindowLongPtr(hWnd, 0))->set_clicked_pixel(LOWORD(lParam) / pixelSize, HIWORD(lParam) / pixelSize);
            InvalidateRect(hWnd, nullptr, false);
            break;

        case WM_KEYDOWN:
            if (wParam == VK_F2 || wParam == VK_F3) {
                if (pixelSize > 1 && wParam == VK_F2) pixelSize--;
                if (pixelSize < 64 && wParam == VK_F3) pixelSize++;

                InvalidateRect(hWnd, nullptr, false);

                char str[256];
                sprintf_s(str, "Масштаб (F2/F3): %d", pixelSize);
                SendMessageA(hWndStatusBar, SB_SETTEXTA, 0, (LPARAM) str);
            }
            if (wParam == VK_F1) {
                MessageBoxA(hWnd, "Работу выполнил студент группы ПВ-221 Колесников А.И.", "О программе", MB_ICONINFORMATION);
            }
            if (wParam == VK_F4) {
                auto newPtr = reinterpret_cast<Painter const *const *>(GetWindowLongPtr(hWnd, sizeof(PainterState *)));
                newPtr++;
                if (newPtr == nullptr)
                    newPtr = &predefined_painters[0];

                SetWindowLongPtr(hWnd, sizeof(PainterState *), (LONG_PTR) newPtr);
            }
            break;

            // Обработка сообщения на изменение размера окна
        case WM_SIZE:
            SendMessageA(hWndStatusBar, WM_SIZE, 0, 0);
            InvalidateRect(hWnd, nullptr, false);
            break;

        case WM_TIMER:
            reinterpret_cast<PainterState *>(GetWindowLongPtr(hWnd, 0))->angle += 0.05;
            InvalidateRect(hWnd, nullptr, false);
            break;


        case WM_DESTROY:
            delete reinterpret_cast<PainterState *>(GetWindowLongPtr(hWnd, 0));
            SetWindowLongPtr(hWnd, 0, reinterpret_cast<LONG_PTR>(nullptr));
            PostQuitMessage(0);
            break;

        default:
            // Все сообщения, не обрабатываемые в данной функции, направляются на обработку по умолчанию
            return DefWindowProcA(hWnd, message, wParam, lParam);
    }

    return 0;
}

