#include <tuple>
#include "frame.hpp"
#include "painter.hpp"


class DemoPainter : public Painter {

    void draw(PainterState *state, Frame *frame) const override {

        int W = frame->width, H = frame->height;
        // Размер рисунка возьмём меньше (7 / 8), чтобы он не касался границ экрана
        float a = 7.0f / 8 * ((W < H) ? W - 1 : H - 1) / sqrt(2);
        if (a < 1) return; // Если окно очень маленькое, то ничего не рисуем
        float angle = state->angle; // Угол поворота
        a = a / 2;

        // Инициализируем исходные координаты центра и вершин квадрата
        struct {
            float x;
            float y;
        } C = {(float) W / 2, (float) H / 2}, A[4] = {{C.x + a, C.y + a},
                                                      {C.x + a, C.y - a},
                                                      {C.x - a, C.y - a},
                                                      {C.x - a, C.y + a}};


        // Поворачиваем все вершины квадрата вокруг точки C на угол angle
        for (int i = 0; i < 4; i++) {
            float xi = A[i].x, yi = A[i].y;
            A[i].x = (xi - C.x) * cos(angle) - (yi - C.y) * sin(angle) + C.x;
            A[i].y = (xi - C.x) * sin(angle) + (yi - C.y) * cos(angle) + C.y;
        }

        // Рисуем стороны квадрата
        for (int i = 0; i < 4; i++) {
            int i2 = (i + 1) % 4;
            frame->draw_line( // Добавляем везде 0.5f, чтобы вещественные числа правильно округлялись при преобразовании к целому типу
                    int(A[i].x + 0.5f),
                    int(A[i].y + 0.5f),
                    int(A[i2].x + 0.5f),
                    int(A[i2].y + 0.5f), COLOR(200, 30, 45));
        }

        // Рисуем описанную окружность
        frame->draw_circle((int) C.x, (int) C.y, int(a * sqrt(2) + 0.5f), COLOR(100, 100, 250));

        // Рисуем пиксель, на который кликнул пользователь
        if (state->clicked_pixel.x >= 0 && state->clicked_pixel.x < W &&
            state->clicked_pixel.y >= 0 && state->clicked_pixel.y < H)
            frame->set_pixel(state->clicked_pixel.x, state->clicked_pixel.y, {34, 175, 60}); // Пиксель зелёного цвета
    }
};

static auto predefined_painters_ = std::make_tuple(
        DemoPainter()
);

Painter const *const predefined_painters[] = {
        &std::get<0>(predefined_painters_),
        nullptr
};